/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault;

import java.io.File;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.management.ObjectName;
import javax.sql.DataSource;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamReader;

import org.apache.axiom.attachments.Attachments;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMException;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.impl.builder.StAXBuilder;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPFaultCode;
import org.apache.axiom.soap.SOAPFaultSubCode;
import org.apache.axiom.soap.SOAPFaultValue;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axiom.soap.impl.builder.MTOMStAXSOAPModelBuilder;
import org.apache.axiom.soap.impl.builder.StAXSOAPModelBuilder;
import org.apache.axiom.soap.impl.dom.soap12.SOAP12Factory;
import org.apache.commons.pool.impl.StackObjectPool;
import org.bluestemsoftware.open.eoa.aspect.axiom.util.STAXUtils;
import org.bluestemsoftware.open.eoa.commons.util.URIUtils;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.axiom.SOAP11FactoryImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.axiom.SOAP12FaultDetailFix;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.mgmt.Binding;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.HeadersUtil;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.ResponseProcessorFactory;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.SOAPFaultContext;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.SOAPReader;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.SOAPWriter;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.dao.MessageDAO;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.dao.MessageModuleDAO;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSReceiver;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSRequestorFactory;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSSenderFactory;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsad.ApplicationDataHandler;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.BindingAction;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFault;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFaultReference;
import org.bluestemsoftware.specification.eoa.component.binding.BindingOperation;
import org.bluestemsoftware.specification.eoa.component.binding.rt.AcceptedResponseError;
import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointOperationReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.ServiceReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule.WSSpecification;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFault;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference.ContentModel;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.MEP;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.BusinessFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.component.service.BindingReference;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.Extension;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryContext;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBinding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingException;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingOperation;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageExchangePattern;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPSystemFaultMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPMustUnderstandFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPReceiverFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPSenderFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPVersionMismatchFault;
import org.bluestemsoftware.specification.eoa.ext.connector.db.DataSourceConnector;
import org.bluestemsoftware.specification.eoa.ext.feature.jms.server.JMSServerFeature;
import org.bluestemsoftware.specification.eoa.ext.feature.jmx.server.JMXServerFeature;
import org.bluestemsoftware.specification.eoa.ext.feature.jmx.server.JMXServerFeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA.WSA10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.WSABindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.WSAD.WSAD10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.soap.ApplicationDataBindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.WSMTOM.WSMTOM10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.soap.WSMTOMBindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.security.soap.WSSEBindingModule;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.bluestemsoftware.specification.eoa.system.container.Container;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;
import org.bluestemsoftware.specification.eoa.system.server.Server;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public abstract class SOAPBindingImpl {

    private static final Log log = SystemContext.getContext().getSystem().getLog(SOAPBinding.class);

    protected SOAPBindingRT consumer;
    protected SOAPBindingConfiguration configuration;
    protected Set<String> features;
    protected SOAPBinding binding;
    protected ExtensionFactoryContext factoryContext;
    protected MessageDAO messageDAO;
    protected MessageModuleDAO messageModuleDAO;
    protected Map<String, EndpointReference> endpointReferences;
    protected Queue receiveQueue;
    protected File dataDirectory;
    protected JMSRequestorFactory jmsRequestorFactory;
    protected Set<JMSReceiver> jmsReceivers;
    protected StackObjectPool jmsRequestors;
    protected StackObjectPool jmsSenders;
    protected StackObjectPool responseProcessorPool;
    protected File tempDirectory;

    private boolean isSuspended = true;
    private QueueConnection receiveConnection;
    private QueueConnection sendConnection;
    private ObjectName objectName;
    private JMXServerFeature jmxServerFeature;

    public SOAPBindingImpl(SOAPBindingConfiguration configuration, Set<String> features) {
        this.configuration = configuration;
        this.endpointReferences = new HashMap<String, EndpointReference>();
        this.features = features;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.UnderlyingProtocolRT$Provider#spi_init()
     */
    public void spi_init() throws SOAPBindingException {

        log.debug("init begin");

        // retrieve static binding component which has same name as runtime
        // binding provider

        binding = (SOAPBinding)consumer.getRuntimeProvidable();

        // initialize dao objects. each will verify that underlying tables exist
        // within default data source, if not, they log a warning and create them.
        // note that all binding instances of this type will work off the same
        // table, implying an AtMostOnce delivery assurance, i.e. without ws-rm

        String connectorName = configuration.getDataSourceConnectorName();
        Container container = SystemContext.getContext().getSystem().getContainer();
        DataSource ds = null;
        try {
            ds = (DataSource)container.getConnector(connectorName);
        } catch (ClassCastException ce) {
            throw new SOAPBindingException("Connector '"
                    + connectorName
                    + "' specified within configuration is not an instance of "
                    + DataSourceConnector.class.getName());
        }
        
        if (ds == null) {
            throw new SOAPBindingException("Connector '"
                    + connectorName
                    + "' specified within configuration is undefined");
        }

        messageDAO = MessageDAO.getInstance(ds).init();
        messageModuleDAO = MessageModuleDAO.getInstance(ds).init();

        // TODO: we need to define a parameter that will allow us to archive
        // the message modules table (message table is insert only, i.e. no
        // performance penalty for size), i.e. archive messages older than
        // x days. i don't think deleting completed meps is appropriate b/c
        // for outgoing messages, the engine cannot log modules via an app
        // module. we're the only one that can do that

        // note that jms server feature is a required feature, i.e. we require
        // that it is embedded. this allows us to stream message content to a
        // file rather than reading entire message into memory (limitation of
        // jms api). and it insures that sender and receiver are both on same
        // machine, i.e. have access to same file system. because connection
        // is not managed, i.e. no transaction or pooling support, we can
        // retrieve directly from feature

        Server server = SystemContext.getContext().getSystem().getServer();
        JMSServerFeature jmsServer = server.getFeature(JMSServerFeature.class);
        ConnectionFactory connFactory = jmsServer.getEmbeddedConnectionFactory(false);

        try {
            receiveConnection = (QueueConnection)connFactory.createConnection();
        } catch (JMSException je) {
            throw new SOAPBindingException(je);
        }

        // construct a unique queue name. we'll use fragment id, because
        // binding qname may clash with other component types defined w/in
        // same namespace

        String queueName = null;
        try {
            URI temp = new URI(binding.getFragmentIdentifier().toString());
            queueName = URIUtils.toFile(temp, null).getName();
            queueName = queueName.replace(".", "_"); // else activemq hangs
            receiveQueue = jmsServer.createQueue(queueName);
        } catch (Exception ex) {
            throw new SOAPBindingException("Error creating receive queue. " + ex);
        }

        // create data directory which we will use to stream message content.
        // this prevents us from having to buffer entire message to memory,
        // i.e. jms api doesn't support handing off inputstream/outpstream

        try {
            dataDirectory = getDataDirectory();
        } catch (Exception ex) {
            throw new SOAPBindingException("Error creating data directory. " + ex);
        }

        // set up our cache of jms objects. note that requestor and sender
        // cache is unconstrained, i.e. because we do not want received
        // actions to wait for a pooled object to become available -> loose
        // coupling

        int minHeaderProcessors = configuration.getMinHeaderProcessors();

        try {
            jmsRequestorFactory = new JMSRequestorFactory(receiveConnection);
            jmsRequestors = new StackObjectPool(jmsRequestorFactory, minHeaderProcessors);
            for (int i = 0; i < minHeaderProcessors; i++) {
                jmsRequestors.addObject();
            }
        } catch (Exception ex) {
            throw new SOAPBindingException(ex);
        }

        try {
            sendConnection = (QueueConnection)connFactory.createConnection();
            JMSSenderFactory jmsSenderFactory = new JMSSenderFactory(sendConnection);
            jmsSenders = new StackObjectPool(jmsSenderFactory, minHeaderProcessors);
            for (int i = 0; i < minHeaderProcessors; i++) {
                jmsSenders.addObject();
            }
            sendConnection.start();
        } catch (Exception ex) {
            throw new SOAPBindingException(ex);
        }

        try {
            int qty = configuration.getMaxPayloadProcessors();
            jmsReceivers = new HashSet<JMSReceiver>(qty);
            for (int i = 0; i < qty; i++) {
                jmsReceivers.add(new JMSReceiver(this, binding.getName(), receiveConnection, receiveQueue,
                        dataDirectory));
            }
        } catch (Exception ex) {
            throw new SOAPBindingException(ex);
        }

        // if jmx server feature is enabled, expose management methods defined
        // on our mbean interface

        jmxServerFeature = server.getFeature(JMXServerFeature.class);
        if (jmxServerFeature != null) {
            String name = null;
            if (getSOAPVersion() == SOAPVersion.SOAP11) {
                name = SOAP11Binding.class.getSimpleName();
            } else {
                name = SOAP12Binding.class.getSimpleName();
            }
            try {
                objectName = jmxServerFeature.registerMBean(consumer, new Binding(this), null, name);
            } catch (JMXServerFeatureException je) {
                throw new SOAPBindingException(je);
            }
        }

        int minThreads = configuration.getMinHeaderProcessors();
        responseProcessorPool = new StackObjectPool(new ResponseProcessorFactory(), minThreads);

        tempDirectory = SystemContext.getContext().getSystem().getSystemTmpDir();

        log.debug("init end");

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT$Provider#spi_getRequiredBindingFeatures()
     */
    public Set<String> spi_getRequiredBindingFeatures() {
        return features;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.Binding$Provider#spi_destroy()
     */
    public void spi_destroy() {
        if (sendConnection != null) {
            try {
                sendConnection.close();
            } catch (JMSException je) {
                log.error("error closing send connection. " + je);
            }
        }
        if (receiveConnection != null) {
            try {
                spi_suspend();
                receiveConnection.close();
            } catch (Exception ex) {
                log.error("error closing receive connection. " + ex);
            }
        }
        binding = null;
        consumer = null;
        if (messageDAO != null) {
            messageDAO.destroy();
        }
        if (messageModuleDAO != null) {
            messageModuleDAO.destroy();
        }
        if (endpointReferences != null) {
            endpointReferences.clear();
        }
        if (objectName != null) {
            try {
                jmxServerFeature.unRegisterMBean(objectName);
            } catch (JMXServerFeatureException ignore) {
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT$Provider#spi_isSuspended()
     */
    public synchronized boolean spi_isSuspended() {
        return isSuspended;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT$Provider#spi_resume()
     */
    public synchronized void spi_resume() throws ExecutableException {
        if (receiveConnection != null) {
            if (!isSuspended) {
                return;
            }
            try {
                receiveConnection.start();
                isSuspended = false;
            } catch (JMSException je) {
                throw new ExecutableException("Error starting receive connection. " + je);
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT$Provider#spi_suspend()
     */
    public synchronized void spi_suspend() throws ExecutableException {
        if (receiveConnection != null) {
            if (isSuspended) {
                return;
            }
            try {
                receiveConnection.stop();
                isSuspended = true;
            } catch (JMSException je) {
                throw new ExecutableException("Error stopping receive connection. " + je);
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.Extension$Provider#spi_setConsumer(org.bluestemsoftware.specification.eoa.ext.Extension.Consumer)
     */
    public void spi_setConsumer(Extension consumer) {
        this.consumer = (SOAPBindingRT)consumer;
        this.factoryContext = consumer.getExtensionFactory().getFactoryContext();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.UnderlyingProtocol$Provider#spi_validateEndpointReference(EndpointReference)
     */
    public void spi_validateEndpointReference(EndpointReference endpointReference) throws DeploymentException {

        // we can't use instance var, i.e. because we may not have been
        // initialized as of yet. retrieve binding via epr

        BindingReference br = endpointReference.getEndpoint().getBindingReference();
        SOAPBinding soapBinding = (SOAPBinding)br.getReferencedComponent();

        // map address to epr object such that we can retrieve component when
        // we deserialize a message instance from message queue. note that if
        // it's a partner service, the callback address is defined via ws-a
        // policy

        ServiceReference serviceReference = (ServiceReference)endpointReference.getParent();
        if (serviceReference.isMyService()) {

            URI address = null;
            try {
                address = new URI(endpointReference.getEndpoint().getAddress());
            } catch (URISyntaxException fatchance) {
            }

            // map address to epr. note that if a callbackaddress is
            // defined for my service it is processed within validation
            // of native binding modules

            endpointReferences.put(address.toString(), endpointReference);

            // account for fact that client could use 'localhost' and or ip
            // address instead of the actual host name. or address may already
            // be localhost (testing) in which case we need to add address
            // with actual hostname

            // derive local ip address which may be used in place of machine
            // name and/or localhost when addressing a message

            String hostName = null;
            String hostAddress = null;
            try {
                hostName = InetAddress.getLocalHost().getHostName();
                hostAddress = InetAddress.getLocalHost().getHostAddress();
            } catch (UnknownHostException ue) {
                throw new DeploymentException(ue);
            }

            String scheme = address.getScheme();
            int port = address.getPort();
            String path = address.getPath();
            try {
                address = new URI(scheme, null, "localhost", port, path, null, null);
                endpointReferences.put(address.toString(), endpointReference);
                address = new URI(scheme, null, hostAddress, port, path, null, null);
                endpointReferences.put(address.toString(), endpointReference);
                address = new URI(scheme, null, hostName, port, path, null, null);
                endpointReferences.put(address.toString(), endpointReference);
            } catch (URISyntaxException fatchance) {
            }

            // if when invoking a service on parent engine and a non-anon
            // response is expected, but instead a soap must understand
            // fault is returned anonymously (as required by spec), we
            // must map this epr to fragment identifier

            String replyToAddress = endpointReference.getFragmentIdentifier().toString();

            if (endpointReferences.get(replyToAddress) == null) {
                endpointReferences.put(replyToAddress, endpointReference);
            }

        } else {
            // defined via ws-addressing policy. entry is added when we process
            // natively implemented binding modules below
        }

        // validate policy defined on natively implemented features. and
        // make sure an installed feature doesn't attempt to implement a
        // feature which we natively implement

        for (EndpointOperationReference eor : endpointReference.getOperationReferences()) {
            for (EndpointActionReference ear : eor.getActionReferences()) {
                for (BindingModule bm : ear.getBindingModules()) {
                    if (((BindingFeature)bm.getFeature()).isNative()) {
                        validateNativeBindingModule(endpointReference, bm);
                    } else {
                        if (bm instanceof WSABindingModule) {
                            throw new DeploymentException("WS-Addressing feature is natively"
                                    + " implemented by this Binding, but BindingFeature"
                                    + " indicates it will implement.");
                        }
                        if (bm instanceof WSMTOMBindingModule) {
                            throw new DeploymentException("WS-MTOM feature is natively"
                                    + " implemented by this Binding, but BindingFeature"
                                    + " indicates it will implement.");
                        }
                        if (bm instanceof WSSEBindingModule) {
                            throw new DeploymentException("WS-Security feature is natively"
                                    + " implemented by this Binding, but BindingFeature"
                                    + " indicates it will implement.");
                        }
                    }
                }
            }
        }

        // while we allow binding to interfaces whose operation utilizes
        // the iri style, we do not support the soap-response MEP, i.e.
        // which would imply that the payload is serialized within the
        // http request URI and http method is GET. soap mep can, however,
        // be request-response, in which case the iri style is supported.
        // it would just constrain schema used to define payload. this
        // would allow user to bind interface with a soap binding and a
        // http binding for RESTful style service invocation

        SOAPMessageExchangePattern sme = null;

        for (BindingOperation sbo : soapBinding.getOperations()) {
            sme = ((SOAPBindingOperation)sbo).getSoapMessageExchangePattern();
            if (sme == null) {
                sme = soapBinding.getSoapMessageExchangePattern();
            }
            if (sme == null) {
                // default mep validated by runtime providable
                sme = SOAPMessageExchangePattern.SOAP_12_REQUEST_RESPONSE;
            }
            if (sme == SOAPMessageExchangePattern.SOAP_12_SOAP_RESPONSE) {
                throw new DeploymentException("SOAP MEP " + sme + " is not supported by this provider.");
            }
        }

        // if we bind more than one interface, i.e. interfaces implemented by
        // referenced services, and soap version is 1.1, robust-in-only mep
        // is not allowed according to wsdl 2.0 soap 1.1 binding spec. we
        // also verify that content model #other is not specified, i.e. this
        // is used for non-xml typed bindings e.g. json

        if (!soapBinding.isReusable()) {
            Service service = serviceReference.getReferencedComponent();
            Interface intrface = service.getInterfaceReference().getReferencedComponent();
            for (InterfaceOperation io : intrface.getOperations()) {
                SOAPVersion soapVersion = soapBinding.getSoapVersion();
                if (io.getMessageExchangePattern().equals(MEP.ROBUST_IN_ONLY)) {
                    if (soapVersion == SOAPVersion.SOAP11) {
                        throw new DeploymentException("SOAP 1.1 Binding "
                                + soapBinding.getName()
                                + " which binds interface "
                                + intrface.getName()
                                + " is invalid. MEP 'robust-in-only' not allowed for SOAP 1.1");
                    }
                }
                InterfaceMessageReference imr = io.getInputMessageReference();
                if (imr != null && imr.getContentModel() != null) {
                    if (imr.getContentModel() == ContentModel.OTHER) {
                        throw new DeploymentException("SOAP Binding "
                                + soapBinding.getName()
                                + " which binds interface "
                                + intrface.getName()
                                + " is invalid. ContentModel #other is not supported.");
                    }
                }
                imr = io.getOutputMessageReference();
                if (imr != null && imr.getContentModel() != null) {
                    if (imr.getContentModel() == ContentModel.OTHER) {
                        throw new DeploymentException("SOAP Binding "
                                + soapBinding.getName()
                                + " which binds interface "
                                + intrface.getName()
                                + " is invalid. ContentModel #other is not supported.");
                    }
                }
            }

            // TODO: once we support RPC style, we need to validate that payload
            // types for input and output conform, i.e. invoke toStyle() method
            // on element decl (after verifying content is modeled as #element)

            // note that if styles exist on interface/operation that we do not
            // use, we simply ignore them. it is responsibility of binding that
            // uses style to validate message def. e.g. it is acceptable for an
            // operation to which we bind to indicate that it uses the iri style.
            // while the element decl should conform to that style, we do not
            // care, i.e. when using this binding, the request would be sent via
            // POST (rather than GET) and we would handle it as normal

        }

    }

    private void validateNativeBindingModule(EndpointReference epr, BindingModule nativeBindingModule) throws DeploymentException {
        if (nativeBindingModule instanceof WSABindingModule) {
            for (WSSpecification spec : nativeBindingModule.getSupportedVersions().values()) {

                if (!spec.getNamespace().equals(WSA10.NAMESPACE_URI)) {
                    throw new DeploymentException("Native binding module defines a version of"
                            + " ws-addressing this binding is not capable of supporting. Found: "
                            + spec.getNamespace()
                            + ".");
                }

                // if a callback address is specified we map the address to
                // epr. replyToAddress is defined via policy with subject
                // endpoint. note that we also process epr's which ref
                // 'my' services, i.e. callback replies returned as result
                // of invoking self (service defined on parent engine)

                WSA10 wsa10 = (WSA10)spec;
                WSA10.Policy policy = (WSA10.Policy)wsa10.getPolicy().iterator().next();

                String replyToAddress = policy.getReplyToAddress();

                if (replyToAddress == null) {

                    // if policy dictates anonymous responses, the callback
                    // address is undefined, i.e. no reason for transport to
                    // install a servlet to handle async responses. we, must,
                    // however, have something to correlate anonymous reply
                    // messages on queue with an epr. so ... we use the
                    // epr fragmentid, which the receivePartnerRequest
                    // method will use to overwrite the anonymous to url

                    replyToAddress = epr.getFragmentIdentifier().toString();

                    if (endpointReferences.get(replyToAddress) == null) {
                        endpointReferences.put(replyToAddress, epr);
                    } else {
                        // multiple response actions with same replyTo
                    }

                } else {

                    URI temp = null;
                    try {
                        temp = new URI(replyToAddress);
                    } catch (URISyntaxException fatchance) {
                    }

                    // derive local ip address which may be used in place of machine
                    // name and/or localhost when addressing a message

                    String hostName = null;
                    String hostAddress = null;
                    try {
                        hostName = InetAddress.getLocalHost().getHostName();
                        hostAddress = InetAddress.getLocalHost().getHostAddress();
                    } catch (UnknownHostException ue) {
                        throw new DeploymentException(ue);
                    }

                    if (endpointReferences.get(temp.toString()) == null) {

                        endpointReferences.put(temp.toString(), epr);

                        // account for fact that client could use 'localhost' and or ip
                        // address instead of the actual host name. or address already
                        // uses localhost (test) in which case we need to add actual
                        // hostname

                        String scheme = temp.getScheme();
                        int port = temp.getPort();
                        String path = temp.getPath();
                        try {
                            temp = new URI(scheme, null, "localhost", port, path, null, null);
                            endpointReferences.put(temp.toString(), epr);
                            temp = new URI(scheme, null, hostAddress, port, path, null, null);
                            endpointReferences.put(temp.toString(), epr);
                            temp = new URI(scheme, null, hostName, port, path, null, null);
                            endpointReferences.put(temp.toString(), epr);
                        } catch (URISyntaxException fatchance) {
                        }

                    } else {

                        // multiple response actions with same replyTo

                    }

                }

            }

        } else if (nativeBindingModule instanceof WSSEBindingModule) {
            throw new DeploymentException("WS-Security feature is not yet implemented");
        } else if (nativeBindingModule instanceof WSMTOMBindingModule) {
            for (WSSpecification spec : nativeBindingModule.getSupportedVersions().values()) {
                if (!spec.getNamespace().equals(WSMTOM10.NAMESPACE_URI)) {
                    throw new DeploymentException("Native binding module defines a version of"
                            + " ws-mtom this binding is not capable of supporting. Found: "
                            + spec.getNamespace()
                            + ".");
                }
            }
        } else {
            throw new DeploymentException("Binding module "
                    + nativeBindingModule.getExtensionType()
                    + " is not a module natively implemented by this Binding.");
        }
    }

    protected SOAPVersion getSOAPVersion() {
        return binding.getSoapVersion();
    }

    protected abstract void sendMyServiceResponse(EndpointReference epr, ActionContext actionContext) throws Exception;

    /**
     * Creates a version specific soap envelope.
     * @param inputStream
     *        stream from which envelope is parsed.
     * @param attachments
     *        or <code>null</code> if media type not 'application/xop+xml'
     * @param charSetEncoding
     *        expected character set encoding
     * 
     * @return soap envelope
     */
    public SOAPEnvelope buildSOAPEnvelope(InputStream inputStream, Attachments attachments, String charSetEncoding) throws SOAPFault {

        SOAPEnvelope envelope;

        try {

            XMLStreamReader reader = STAXUtils.createXMLStreamReader(inputStream, charSetEncoding);

            StAXBuilder builder = null;
            if (attachments == null) {
                builder = new StAXSOAPModelBuilder(reader, getSOAPFactory(), null);
            } else {
                builder = new MTOMStAXSOAPModelBuilder(reader, getSOAPFactory(), attachments, null);
            }

            envelope = (SOAPEnvelope)builder.getDocumentElement();

            String soapEnvelopeNamespace = envelope.getNamespace().getNamespaceURI();

            // the version passed implies namespace of fault code, which should
            // match envelope version received, not the version expected

            if (getSOAPVersion() == SOAPVersion.SOAP11) {
                if (!soapEnvelopeNamespace.equals(Constants.SOAP_11_ENVELOPE_NS)) {
                    throw new SOAPVersionMismatchFault(SOAPVersion.SOAP12);
                }
            } else {
                if (!soapEnvelopeNamespace.equals(Constants.SOAP_12_ENVELOPE_NS)) {
                    throw new SOAPVersionMismatchFault(SOAPVersion.SOAP11);
                }
            }

            String xmlCharSetEncoding = builder.getDocument().getCharsetEncoding();
            if (xmlCharSetEncoding != null && !xmlCharSetEncoding.equals("")) {
                if (!xmlCharSetEncoding.equals(charSetEncoding)) {
                    throw new SOAPSenderFault(getSOAPVersion(), "Character set encoding defined on xml decl '"
                            + xmlCharSetEncoding
                            + "' failed to match expected encoding '"
                            + charSetEncoding
                            + "'.");
                }
            }

            checkEnvelopeStructure(getSOAPVersion(), envelope);

        } catch (OMException oe) {

            Throwable th = oe.getCause();
            if (th != null && th instanceof com.ctc.wstx.exc.WstxEOFException) {
                return null;
            }

            // this is more than likely due to a malformed soap request, i.e. throw
            // a soap sender fault.

            throw new SOAPSenderFault(getSOAPVersion(), "The request was malformed. " + oe.getMessage());

        } catch (Throwable th) {

            if (th instanceof SOAPFault) {
                throw (SOAPFault)th;
            }

            // this is more than likely due to a problem on our end, i.e. throw a
            // soap receiver fault

            if (configuration.isEmbeddStackTraceInFaults()) {
                throw new SOAPReceiverFault(getSOAPVersion(), "Error reading message. ", th);
            } else {
                log.warn("Error reading message. ", th);
                throw new SOAPReceiverFault(getSOAPVersion(), "Error reading message. " + th.getMessage());
            }

        }

        return envelope;

    }

    /**
     * Creates an empty version specific soap envelope.
     * @return
     */
    public SOAPEnvelope buildSOAPEnvelope() {
        if (getSOAPVersion() == SOAPVersion.SOAP11) {
            return getSOAPFactory().getDefaultEnvelope();
        } else {
            return getSOAPFactory().getDefaultEnvelope();
        }
    }

    /**
     * Creates a version specific soap fault in accordance with binding rules for action
     * defined on soap fault, which may be undeclared.
     * 
     * @param soapFault
     * @return
     */
    public SOAPEnvelope buildSOAPFaultEnvelope(EndpointReference endpointReference, SOAPFault soapFault) {

        // if it's a soap version mismatch, we use soap version defined
        // within fault to determine version of envelope returned, else
        // we use binding version

        SOAPVersion soapVersion = null;
        if (soapFault instanceof SOAPVersionMismatchFault) {
            soapVersion = ((SOAPVersionMismatchFault)soapFault).getSoapVersion();
        } else {
            soapVersion = getSOAPVersion();
        }
        if (soapVersion == SOAPVersion.SOAP11) {
            return buildSOAP11FaultEnvelope(endpointReference, soapFault);
        } else {
            return buildSOAP12FaultEnvelope(endpointReference, soapFault);
        }

    }

    /**
     * Binds message by decoding wire formatted message into an abstract message instance
     * according to <code>Binding</code> type specific rules and passes result to
     * <code>Engine</code>. This method is invoked by instances of {@link JMSReceiver}.
     * 
     * @param messageID
     * @param myService
     *        required in addition to messageid to make primary key unique, i.e. when partner
     *        engine is defined on same machine. true if receiving my service request or
     *        sending my service response. false if sending partner service request or
     *        receiving partner service response.
     * @param inputStream
     *        soap body content or <code>null</code> if ##none
     * @return runtime abstract message instance
     * @throws org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFault
     */
    public void readMessage(String messageID, boolean myService, InputStream inputStream) throws org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT {
        log.trace("readMessage begin");

        EndpointReference epr = null;
        EndpointActionReference ear = null;
        ActionContext actionContext = null;
        MessageModules messageModules = null;
        WSAMessageModule wsamm = null;
        Message message = null;
        Direction direction = null;

        try {

            messageModules = messageModuleDAO.selectMessageModules(messageID, myService);
            wsamm = messageModules.getModule(WSAMessageModule.class);

            epr = endpointReferences.get(wsamm.getTo());
            ear = epr.getEndpointActionReference(wsamm.getAction());

            InterfaceMessage am = null;
            SOAPMessage cm = null;

            // if we're processing a partner response, we need to allow for the
            // fact that it may be a fault. note that caller insures that action
            // is declared. all undeclared actions are converted to a sys fault
            // message

            QName faultName = null;
            if (ear != null) {
                InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();
                direction = ia.getDirection();
                if (ia instanceof InterfaceMessageReference) {
                    am = ((InterfaceMessageReference)ia).getReferencedComponent();
                } else {
                    InterfaceFault interfaceFault = ((InterfaceFaultReference)ia).getFault();
                    am = interfaceFault.getReferencedComponent();
                    faultName = interfaceFault.getName();
                }
                cm = getConcreteMessageDefinition(ear.getAction());
            } else {
                FragmentIdentifier fragmentID = SystemFault.MESSAGE_FRAGMENT_ID;
                am = (InterfaceMessage)SystemContext.getContext().getSystem().getComponent(fragmentID);
                cm = SOAPSystemFaultMessage.getInstance();
                direction = Direction.OUT;
            }

            message = SOAPReader.readMessage(getSOAPVersion(), epr, am, cm, messageModules, inputStream);

            if (ear == null) {
                ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();
                actionContext = new FaultContext(messageModules, SystemFault.valueOf(app, message));
            } else {
                if (faultName == null) {
                    actionContext = new MessageContext(messageModules, message);
                } else {
                    actionContext = new FaultContext(messageModules, new BusinessFault(faultName, message));
                }
            }

        } catch (Throwable th) {

            // soap reader should only throw soap sender and/or soap receiver
            // faults. anything else is a programming error that should be
            // logged

            if (th instanceof SOAPFault) {
                if (log.isDebugEnabled()) {
                    log.debug("caught soap fault while binding received action "
                            + wsamm.getAction()
                            + " defined on endpoint referenced by component "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SOAPFault)th).getFaultReason());
                }
            } else {
                log.error("caught throwable while binding received action "
                        + wsamm.getAction()
                        + " defined on endpoint referenced by component "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + th);
            }

            if (direction == Direction.IN) {

                // we encountered an error while receiving a request targeted to
                // locally hosted service. if mep is in-out or robust-in-only,
                // we return a response to client, i.e. request failed within
                // binding layer and will not be passed to engine

                EndpointOperationReference eor = (EndpointOperationReference)ear.getParent();
                MEP mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

                if (mep == MEP.IN_ONLY) {

                    // basic profile requires an empty body in response to one way mep,
                    // with an accepted code, even if we failed to process the request

                    log.debug("MEP is in-only. No fault response allowed.");
                    return;

                }

                SOAPFault sf = null;
                if (th instanceof SOAPFault) {
                    sf = (SOAPFault)th;
                } else {
                    if (configuration.isEmbeddStackTraceInFaults()) {
                        sf = new SOAPReceiverFault(getSOAPVersion(), th);
                    } else {
                        sf = new SOAPReceiverFault(getSOAPVersion(), th.getMessage());
                    }
                }

                // pass response to sendMyServiceResponse method which will handle
                // module processing, etc ... and will manage sync/async transport

                try {
                    WSAMessageModule wsammOut = new WSAMessageModule(sf.getAction());
                    wsammOut.setRelatesTo(wsamm.getMessageID());
                    MessageModules responseModules = new MessageModules(wsammOut);
                    SOAPFaultContext sfc = new SOAPFaultContext(null, messageModules, responseModules, sf);
                    sendMyServiceResponse(epr, sfc);
                } catch (Throwable thr) {
                    log.error("caught throwable while returning soap fault response. " + thr);
                }

                return;

            } else {

                // we're processing a partner response. convert fault to an instance
                // of accepted response fault containing partner response, if avail,
                // and then convert to a system fault, which will be passed to app

                QName faultCode = null;
                String faultReason = null;

                if (th instanceof SOAPFault) {
                    faultCode = ((SOAPFault)th).getFaultCode();
                    faultReason = ((SOAPFault)th).getFaultReason();
                } else {
                    faultReason = "Caught throwable while processing partner response. " + th.toString();
                }

                List<Content> payload = null;
                if (th instanceof SOAPSenderFault) {
                    payload = Arrays.asList(((SOAPSenderFault)th).getPayload());
                }

                AcceptedResponseError are = new AcceptedResponseError(faultCode, faultReason, payload);
                ApplicationRT app = ((EngineRT)ear.getRootComponent()).getApplication();
                SystemFault systemFault = SystemFault.valueOf(app, are);
                actionContext = new FaultContext(messageModules, systemFault);

            }

        }

        // message has successfully traversed through the 'binding layer'. now we
        // pass the abstract modules and abstract message to engine which is
        // responsible for guiding it through the 'application layer'

        try {

            if (log.isDebugEnabled()) {
                log.debug("invoking receiveAction on epr " + epr.getFragmentIdentifier().toString());
            }

            epr.receiveAction(actionContext);

            log.debug("action received");

        } catch (Throwable th) {

            // engine should not propogate exceptions. all we can do here is log
            // an error, i.e. if this is a partner svc response, invoking engine
            // again would probably cause same error. if this is a my service
            // request, engine may have successfully processed the request and
            // allowed an exception, i.e. a system fault, to propogate from the
            // sendAction method while returning response - attempting to unwravel
            // this would be ugly at best

            log.error("engine threw/propogated an exception while receiving action "
                    + wsamm.getAction()
                    + " defined on endpoint referenced by component "
                    + epr.getFragmentIdentifier()
                    + ". "
                    + th);

            // note that if this is a my service anonymous request, the jms
            // requestor will eventually timeout and return a timeout error
            // message to client (as a soap fault)

        }

        log.trace("readMessage end");
    }

    /**
     * Binds a normal message, i.e. an input or output message by encoding an abstract message
     * instance into a concrete message instance according to <code>Binding</code> type
     * specific rules.
     * 
     * @param endpointActionReference
     *        statically defines referenced action with which message is associated
     * @param messageContext
     *        which contains abstract message instance
     * 
     * @return concrete message instance
     * @throws SystemFault
     *         if an error occurs encoding message or if abstract message fails to comply with
     *         structure statically defined by corresponding {@link BindingAction}.
     */
    protected SOAPEnvelope writeMessage(EndpointReference epr, MessageContext messageContext) throws SystemFault {
        log.trace("writeMessage begin");
        SOAPEnvelope soapEnvelope = buildSOAPEnvelope();
        EndpointActionReference ear = epr.getEndpointActionReference(messageContext.getAction());
        InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();
        InterfaceMessage am = ((InterfaceMessageReference)ia).getReferencedComponent();
        SOAPMessage cm = getConcreteMessageDefinition(ear.getAction());
        SOAPWriter.writeMessage(ear, am, cm, getSOAPVersion(), messageContext, soapEnvelope);
        log.trace("writeMessage end");
        return soapEnvelope;
    }

    /**
     * Extracts application data from soap headers as required by 'Application Data Feature.'
     * 
     * @param ear
     * @param soapEnvelope
     * @return
     */
    protected Map<QName, Set<ApplicationData>> readApplicationData(EndpointActionReference ear, SOAPEnvelope soapEnvelope) {

        // retrieve this list of headers 'understood' by targeted service
        // defined via wsad policy

        ServiceReference sref = ear.getParent().getParent().getParent();
        Set<QName> uh = null;
        ApplicationDataBindingModule adbm = ear.getBindingModule(ApplicationDataBindingModule.class);
        if (adbm != null) {
            WSSpecification spec = adbm.getSupportedVersions().get(WSAD10.NAMESPACE_URI);
            WSAD10.Policy policy = (WSAD10.Policy)spec.getPolicy().iterator().next();
            uh = policy.getUnderstoodHeaders();
        } else {
            uh = new HashSet<QName>();
        }

        if (sref.isMyService()) {

            // if this is a request addressed to my service, then we use the
            // role name as is to identify target service

            String roleName = sref.getRoleName();
            return ApplicationDataHandler.handleMyServiceRequest(roleName, uh, soapEnvelope);

        } else {

            // if we are processing a partner response, we use the callback
            // role name

            String roleName = sref.getConsumerRoleName();
            return ApplicationDataHandler.handlePartnerServiceResponse(roleName, uh, soapEnvelope);

        }

    }

    /**
     * Writes application data as soap headers as required by 'Application Data Feature.' Note
     * that we do not use SOAPMessage definition when writing headers. If sent message is non
     * compliant, it is the responsibility of receiver to throw a fault if a required header is
     * missing.
     * 
     * @param ear
     * @param messageModules
     * @param soapEnvelope
     */
    protected void writeApplicationData(EndpointActionReference ear, MessageModules messageModules, SOAPEnvelope soapEnvelope) {
        ServiceReference sref = ear.getParent().getParent().getParent();
        if (sref.isMyService()) {
            ApplicationDataHandler.handleMyServiceResponse(messageModules, soapEnvelope);
        } else {
            ApplicationDataHandler.handlePartnerServiceRequest(messageModules, soapEnvelope);
        }
    }
 
    /**
     * Examines soap header blocks checking for required headers as declared on concrete
     * message definition and throws a soap fault if any are missing.
     * 
     * @param message
     * @throws SOAPFault
     */
    protected void checkRequiredSOAPHeaders(SOAPMessage soapMessage, SOAPEnvelope message) throws SOAPFault {

        if (soapMessage == null || soapMessage.getRequiredHeaders() == null) {
            return;
        }

        // note that according to basic profile (R2724) 'in all other cases
        // where an envelope is inconsistent with its WSDL description, a
        // fault with a 'Client' faultcode should be generated'. so, we'll
        // throw soap sender faults

        Set<QName> requiredHeaders = soapMessage.getRequiredHeaders();

        if (message.getHeader() == null) {
            QName requiredHeader = requiredHeaders.iterator().next();
            throw new SOAPSenderFault(getSOAPVersion(), "Required header " + requiredHeader + " is missing.");
        }

        Set<QName> foundHeaders = new HashSet<QName>();

        OMNode node = message.getHeader().getFirstElement();
        while (node != null && node.getType() == OMNode.ELEMENT_NODE) {
            OMNamespace namespace = ((OMElement)node).getNamespace();
            String namespaceURI = namespace == null ? null : namespace.getNamespaceURI();
            foundHeaders.add(new QName(namespaceURI, ((OMElement)node).getLocalName()));
            node = node.getNextOMSibling();
        }

        // if a required header is missing, throw a fault. note that by this
        // point, ws-addressing is engaged, which implies ws-headers have
        // already been extracted

        for (QName requiredHeader : requiredHeaders) {
            String namespaceURI = requiredHeader.getNamespaceURI();
            if (!foundHeaders.contains(requiredHeader)) {
                if (namespaceURI != null && namespaceURI.equals(WSA10.NAMESPACE_URI)) {
                    continue;
                }
                throw new SOAPSenderFault(getSOAPVersion(), "Required header " + requiredHeader + " is missing.");
            }
        }

    }

    /**
     * Examines remaining soap header blocks and throws a must understand fault if any exist
     * that we do not understand.
     * 
     * @param message
     * @throws SOAPFault
     */
    protected void checkMustUnderstand(SOAPEnvelope message) throws SOAPFault {

        SOAPHeader header = message.getHeader();
        if (header == null) {
            return;
        }

        // because modules are supposed to extract a header after it has
        // been processed, if any header exists that's intended for
        // ultimate destination, i.e. that's us, and is flagged as must
        // understand, this is an error

        Iterator<OMElement> itr = HeadersUtil.getHeadersForUltimateDestination(getSOAPVersion(), header, null);

        Set<QName> opaqueHeaders = null;
        while (itr.hasNext()) {
            OMElement hb = itr.next();
            String value = null;
            if (getSOAPVersion() == SOAPVersion.SOAP11) {
                value = hb.getAttributeValue(Constants.SOAP_11_MUST_UNDERSTAND_ATTRIB);
            } else {
                value = hb.getAttributeValue(Constants.SOAP_12_MUST_UNDERSTAND_ATTRIB);
            }
            boolean mustUnderstand = false;
            if (value != null) {
                value = value.equals("1") ? "true" : value.equals("0") ? "false" : value;
                mustUnderstand = Boolean.parseBoolean(value);
            }
            if (mustUnderstand) {
                if (opaqueHeaders == null) {
                    opaqueHeaders = new HashSet<QName>();
                }
                opaqueHeaders.add(hb.getQName());
            }
        }

        if (opaqueHeaders != null) {
            throw new SOAPMustUnderstandFault(getSOAPVersion(), opaqueHeaders);
        }

    }

    /**
     * Retreives concrete message def for indicated action.
     * 
     * @param action
     * @return SOAPMessage or <code>null</code> if action not bound
     */
    protected SOAPMessage getConcreteMessageDefinition(String action) {
        BindingAction ba = binding.getBindingAction(action);
        if (ba != null) {
            if (ba instanceof SOAPBindingMessageReference) {
                return ((SOAPBindingMessageReference)ba).getSoapMessage();
            } else {
                BindingFault bf = ((BindingFaultReference)ba).getFault();
                return ((SOAPBindingFault)bf).getSoapMessage();
            }
        } else {
            return null;
        }
    }

    private SOAPEnvelope buildSOAP11FaultEnvelope(EndpointReference endpointReference, SOAPFault soapFault) {

        if (soapFault.getSoapVersion() != SOAPVersion.SOAP11) {
            throw new IllegalArgumentException("Expected a SOAP 1.1 fault");
        }

        SOAPFactory factory = new SOAP11FactoryImpl();
        SOAPEnvelope envelope = factory.getDefaultFaultEnvelope();
        org.apache.axiom.soap.SOAPFault sf = envelope.getBody().getFault();
        SOAPFaultCode soapFaultCode = sf.getCode();
        if (soapFaultCode.getFirstElement() != null) {
            // detach soap 11 value element which we do not use
            // and which could cause inconsistencies when using
            // dom to retrieve child text node
            soapFaultCode.getFirstElement().detach();
        }
        QName faultCode = soapFault.getFaultCode();
        if (faultCode.getNamespaceURI() != null) {
            OMNamespace ns = soapFaultCode.findNamespace(faultCode.getNamespaceURI(), null);
            if (ns == null) {
                soapFaultCode.declareNamespace(faultCode.getNamespaceURI(), "ns");
                soapFaultCode.setText("ns:" + faultCode.getLocalPart());
            } else {
                soapFaultCode.setText(ns.getPrefix() + ":" + faultCode.getLocalPart());
            }
        } else {
            soapFaultCode.setText(faultCode.getLocalPart());
        }

        if (sf.getReason().getFirstElement() != null) {
            // detach soap 11 reason text element which we do not
            // use and which could cause inconsistencies when using
            // dom to retrieve child text node
            sf.getReason().getFirstElement().detach();
        }

        sf.getReason().setText(soapFault.getFaultReason());

        // fault is responsible for appending detail to the appropriate
        // element which could be detail element or header, i.e. when
        // soap 1.1 and fault not related to body of request

        soapFault.appendFaultDetail((Content)envelope);

        return envelope;

    }

    private SOAPEnvelope buildSOAP12FaultEnvelope(EndpointReference endpointReference, SOAPFault soapFault) {

        if (soapFault.getSoapVersion() != SOAPVersion.SOAP12) {
            throw new IllegalArgumentException("Expected a SOAP 1.2 fault");
        }

        SOAPFactory soapFactory = new SOAP12Factory();
        SOAPEnvelope soapEnvelope = soapFactory.getDefaultFaultEnvelope();

        if (soapFault instanceof SOAPMustUnderstandFault) {
            SOAPHeader header = soapEnvelope.getHeader();
            org.apache.axiom.soap.SOAPHeaderBlock shb = null;
            for (QName name : ((SOAPMustUnderstandFault)soapFault).getHeaderNames()) {
                shb = header.addHeaderBlock(Constants.SOAP_NOT_UNDERSTOOD, soapEnvelope.getNamespace());
                shb.addAttribute("qname", "ns:" + name.getLocalPart(), null);
                shb.declareNamespace(name.getNamespaceURI(), "ns");
            }
        }

        QName faultCode = null;
        List<QName> faultSubcodes = null;

        EndpointActionReference ear = endpointReference.getEndpointActionReference(soapFault.getAction());

        if (ear != null) {
            SOAPBindingFault sbf = (SOAPBindingFault)binding.getFault(soapFault.getFaultName());
            if (sbf != null) {
                if (sbf.getSoapFaultCode() == null) {
                    faultCode = soapFault.getFaultCode();
                } else {
                    // should be #any or soap:Receiver, but user may change
                    faultCode = sbf.getSoapFaultCode();
                }
                if (sbf.getSoapFaultSubCodes().size() == 0) {
                    faultSubcodes = soapFault.getFaultSubCodes();
                } else {
                    // should be #any or faultName, but user may change
                    faultSubcodes = sbf.getSoapFaultSubCodes();
                }
            } else {
                faultCode = soapFault.getFaultCode();
                faultSubcodes = soapFault.getFaultSubCodes();
            }
        } else {

            faultCode = soapFault.getFaultCode();
            faultSubcodes = soapFault.getFaultSubCodes();

        }

        org.apache.axiom.soap.SOAPFault sf = soapEnvelope.getBody().getFault();
        SOAPFaultCode soapFaultCode = sf.getCode();
        SOAPFaultValue sfv = soapFaultCode.getValue();
        if (sfv == null) {
            sfv = soapFactory.createSOAPFaultValue(soapFaultCode);
        }
        if (faultCode.getNamespaceURI() != null) {
            OMNamespace ns = sfv.findNamespace(faultCode.getNamespaceURI(), null);
            if (ns == null) {
                sfv.declareNamespace(faultCode.getNamespaceURI(), "ns");
                sfv.setText("ns:" + faultCode.getLocalPart());
            } else {
                sfv.setText(ns.getPrefix() + ":" + faultCode.getLocalPart());
            }
        } else {
            sfv.setText(faultCode.getLocalPart());
        }

        SOAPFaultSubCode soapFaultSubCode = null;
        for (QName faultSubcode : faultSubcodes) {
            if (soapFaultSubCode == null) {
                soapFaultSubCode = soapFactory.createSOAPFaultSubCode(soapFaultCode);
            } else {
                soapFaultSubCode = soapFactory.createSOAPFaultSubCode(soapFaultSubCode);
            }
            sfv = soapFactory.createSOAPFaultValue(soapFaultSubCode);
            if (faultSubcode.getNamespaceURI() != null) {
                OMNamespace ns = sfv.findNamespace(faultSubcode.getNamespaceURI(), null);
                if (ns == null) {
                    sfv.declareNamespace(faultSubcode.getNamespaceURI(), "ns");
                    sfv.setText("ns:" + faultSubcode.getLocalPart());
                } else {
                    sfv.setText(ns.getPrefix() + ":" + faultSubcode.getLocalPart());
                }
            } else {
                sfv.setText(faultSubcode.getLocalPart());
            }
        }

        sf.getReason().getFirstSOAPText().setText(soapFault.getFaultReason());

        Document owner = ((Element)sf.getDetail().detach()).getOwnerDocument();
        sf.addChild(new SOAP12FaultDetailFix(owner, soapFactory));

        soapFault.appendFaultDetail((Content)soapEnvelope);

        return soapEnvelope;

    }

    private File getDataDirectory() throws Exception {
        File systemVarDir = SystemContext.getContext().getSystem().getSystemVarDir();
        File temp = URIUtils.toFile(new URI(binding.getFragmentIdentifier().toString()), null);
        File result = new File(systemVarDir, temp.toString());
        if (!result.exists()) {
            result.mkdir();
        }
        return result;
    }

    private SOAPFactory getSOAPFactory() {

        // create a new instance of dom factory per invocation. DOMSOAPFactory
        // is lightweight and ... it creates a single document which it saves
        // to an instance var, i.e. the document would be shared across all
        // threads, which could be problematic (as of version 1.2.5)

        if (getSOAPVersion() == SOAPVersion.SOAP11) {
            return new SOAP11FactoryImpl();
        } else {
            return new SOAP12Factory();
        }

    }

    /*
     * As of version 1.2.5 snapshot, axiom does not do a good job of checking envelope
     * structure
     */
    private void checkEnvelopeStructure(SOAPVersion sv, OMElement soapEnvelope) throws SOAPSenderFault {

        OMElement firstElement = soapEnvelope.getFirstElement();

        if (firstElement instanceof SOAPHeader) {
            if (sv == SOAPVersion.SOAP11) {
                if (!firstElement.getNamespace().getNamespaceURI().equals(Constants.SOAP_11_ENVELOPE_NS)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Header namespace not "
                            + Constants.SOAP_11_ENVELOPE_NS
                            + ".");
                }
            } else {
                if (!firstElement.getNamespace().getNamespaceURI().equals(Constants.SOAP_12_ENVELOPE_NS)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Header namespace not "
                            + Constants.SOAP_12_ENVELOPE_NS
                            + ".");
                }
            }
            OMNode node = firstElement.getNextOMSibling();
            while (node != null && node.getType() != OMNode.ELEMENT_NODE) {
                node = node.getNextOMSibling();
            }
            OMElement siblingElement = (OMElement)node;
            if (siblingElement == null) {
                throw new SOAPSenderFault(sv,
                        "Malformed request. SOAP envelope missing required soap:Body element.");
            }
            if (sv == SOAPVersion.SOAP11) {
                if (!siblingElement.getQName().equals(Constants.SOAP_11_BODY_TYPE)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Header sibling not soap:Body element.");
                }
            } else {
                if (!siblingElement.getQName().equals(Constants.SOAP_12_BODY_TYPE)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Header sibling not soap:Body element.");
                }
            }
            // note that we cannot parse beyond root of body, i.e. to avoid consuming
            // entire body to memory.
        } else if (firstElement instanceof SOAPBody) {
            if (sv == SOAPVersion.SOAP11) {
                if (!firstElement.getNamespace().getNamespaceURI().equals(Constants.SOAP_11_ENVELOPE_NS)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Body namespace not "
                            + Constants.SOAP_11_ENVELOPE_NS
                            + ".");
                }
            } else {
                if (!firstElement.getNamespace().getNamespaceURI().equals(Constants.SOAP_12_ENVELOPE_NS)) {
                    throw new SOAPSenderFault(sv, "Malformed request. soap:Body namespace not "
                            + Constants.SOAP_12_ENVELOPE_NS
                            + ".");
                }
            }
            // note that we cannot parse beyond root of body, i.e. to avoid consuming
            // entire body to memory.
        }

    }

}
