/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util;


/**
 * Returns binding fault response on separate thread, i.e. separate from thread that processed
 * request. Allows request thread to set set http response status to 'accepted' and return
 * immediately.
 */
public class ResponseProcessor extends Thread {

    Work work;
    Object lock = new Object();

    public ResponseProcessor() {
        super("BindingFaultResponseProcessor");
    }

    public void setWork(Work work) {
        synchronized (lock) {
            this.work = work;
            lock.notify();
        }
    }

    @Override
    public void run() {
        while (!Thread.interrupted()) {
            synchronized (lock) {
                while (work == null) {
                    try {
                        lock.wait();
                    } catch (InterruptedException ie) {
                        break;
                    }
                }
                work.doWork();
                work = null;
            }
        }
    }
    
    public interface Work {
        
        public void doWork();
        
    }

}
