/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsse;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPHeader;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.HeaderRequiredFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.security.WSSEMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.security.WSSE.WSSE11;

public abstract class WSSEHandler {

    /**
     * Gets instance which implements rules defined within verion implied by indicated
     * namespace.
     * @param namespace
     * @return
     */
    public static WSSEHandler getInstance(String namespace) {
        if (namespace.equals(WSSE11.NAMESPACE_URI)) {
            return WSSE11Handler.getInstance();
        } else {
            throw new IllegalArgumentException("Namespace not supported");
        }
    }

    /**
     * Gets instance based upon reliable messaging namespace used within headers defined on a
     * request received on behalf of my service, or a response recieved from a partner service.
     * 
     * @param message
     * @return
     * @throws SOAPFault
     */
    public static WSSEHandler getInstance(SOAPEnvelope message) throws SOAPFault {

        SOAPHeader soapHeader = message.getHeader();

        if (soapHeader == null) {
            String envelopeNamespace = message.getNamespace().getNamespaceURI();
            SOAPVersion soapVersion = SOAPVersion.valueOfEnvelopeNamespace(envelopeNamespace);
            throw new HeaderRequiredFault(soapVersion, WSSE11.NAMESPACE_URI, WSA.ACTION);
        }

        String matchingNamespace = null;

        OMNode node = soapHeader.getFirstElement();

        while (node != null && node.getType() == OMNode.ELEMENT_NODE) {
            OMNamespace namespace = ((OMElement)node).getNamespace();
            if (namespace == null) {
                continue;
            }
            if (namespace.getNamespaceURI().equals(WSSE11.NAMESPACE_URI)) {
                matchingNamespace = WSSE11.NAMESPACE_URI;
                break;
            }
            node = node.getNextOMSibling();
        }

        if (matchingNamespace == null) {
            // TODO: what gets thrown here?
        }

        return WSSEHandler.getInstance(matchingNamespace);

    }

    public abstract WSSEMessageModule handleMyServiceRequest(EndpointReference epr, String action, SOAPEnvelope message) throws BindingFaultRT;

    public abstract WSSEMessageModule handleMyServiceResponse(MessageModules requestModules, MessageModules responseModules, SOAPEnvelope message) throws BindingFaultRT;

    public abstract WSSEMessageModule handlePartnerServiceRequest(MessageModules requestModules, SOAPEnvelope message) throws BindingFaultRT;

    public abstract WSSEMessageModule handlePartnerServiceResponse(EndpointReference epr, String action, SOAPEnvelope message) throws BindingFaultRT;

}
